from .message import (
    make_message,
    BaseMessage,
    MinimalMessage,
    MetadataMessage,
    FullMessage,
    RawMessage
)
from .thread import Thread
from .label import Label
from .history import History
from .draft import Draft
from . import *

from ._base_object import _BaseObject


from typing import(
    Callable
)

def override(interfaceClass: type):
    def overrider(method: Callable):
        try:
            assert(method.__name__ in dir(interfaceClass))
        except AttributeError:
            pass
        return method
    return overrider


from typing import (
    List,
    Dict,
    Optional
)
from abc import (
    ABC,
    abstractproperty
)
import json
import email
from datetime import datetime
from .._base_object import _BaseObject


class BaseMessage(_BaseObject):
    
    def __init__(self,
                 id: str,
                 threadId: str,
                 labelIds: List[str],
                 snippet: str,
                 historyId: str,
                 internalDate: str,
                 sizeEstimate: int):
        super().__init__(id)
        self._thread_id = threadId
        self._label_ids = labelIds
        self._snippet = snippet
        self._history_id = historyId
        self._internal_date = int(internalDate)
        self._size_estimate = sizeEstimate

    @property
    def email_message(self) -> email.message.Message:
        return email.message.Message()

    @property
    def thread_id(self) -> str:
        return self._thread_id

    @property
    def label_ids(self) -> List[str]:
        return self._label_ids

    @property
    def sent(self) -> bool:
        return 'SENT' in self._label_ids
    
    @property
    def snippet(self) -> str:
        return self._snippet

    @property
    def history_id(self) -> str:
        return self._history_id

    @property
    def internal_date(self) -> int:
        return self._internal_date
    
    @property
    def datetime(self) -> datetime:
        return datetime.fromtimestamp(self.internal_date/1000)

    @property
    def size_estimate(self) -> int:
        return self._size_estimate



from typing import(
    Iterable,
    Dict,
    Any,
    Union,
    Optional
)
#
import json
from datetime import datetime
import azure.common
import azure.core.exceptions
from azure.storage.blob import (
    BlobServiceClient
)
from azure.storage. filedatalake import (
    DataLakeServiceClient
)
from azure.storage.file import (
    FileService
)
from azure.cosmosdb.table import (
    TableService,
    Entity
)
#
from npyetl.internals.decorators import override
from npyetl.data import BaseDataBlock
from npyetl.load import BaseLoader


class LoaderAzureBlob(BaseLoader):
    """
    Loader for Azure Blob Storage files.
    """
    def __init__(self,
                 version: int = 2,
                 metadata: dict = None):
        super().__init__(metadata)
    
    @override(BaseLoader)
    def load(self,
             data_blocks: Iterable[BaseDataBlock],
             storage_account_name: str,
             storage_account_key: str,
             container_name: str,
             directory: str = '',
             overwrite: bool = True,
             upload_blob_kwargs: dict = None,
             extension: str = None,
             **kwargs) -> bool:
        """
        Creates, updates or overwrites a blob file in the given storage account.
        :param storage_account_name:
        :param storage_account_key:
        :param container_name:
        :param directory:
        :param overwrite:
        :param upload_blob_kwargs:
        """
        upload_blob_kwargs = upload_blob_kwargs or dict()
        blob_storage_url = 'https://%s.blob.core.windows.net/' % storage_account_name
        blob_service_client = BlobServiceClient(blob_storage_url, credential=storage_account_key)
        container_client = blob_service_client.get_container_client(container_name)
        for data_block in data_blocks:
            blob_path = directory + data_block.name + (extension if extension else '')
            blob_client = container_client.get_blob_client(blob_path)
            try:
                blob_client.upload_blob(data_block.__str__(), **upload_blob_kwargs)
            except azure.core.exceptions.ResourceExistsError:
                if overwrite:
                    blob_client.delete_blob()
                    blob_client.upload_blob(data_block.__str__(), **upload_blob_kwargs)
        return True


class LoaderAzureTable(BaseLoader):
    """
    Loader for Azure Table Storage.
    """
    def __init__(self,
                 version: int = 2,
                 metadata: dict = None):
        super().__init__(metadata)
    
    @override(BaseLoader)
    def load(self,
             data_blocks: Iterable[BaseDataBlock],
             storage_account_name: str,
             storage_account_key: str,
             table_name: str,
             partition_key: Optional[str] = None,
             row_key: Optional[str] = None,
             overwrite: bool = True,
             timeout: int = None,
             **kwargs) -> bool:
        """
        :param storage_account_name:
        :param storage_account_key:
        :param container_name:
        :param directory:
        :param overwrite:
        :param upload_blob_kwargs:
        """
        table_service = TableService(storage_account_name, storage_account_key)
        for data_block in data_blocks:
            entity = self._set_entity(data_block.records, partition_key, row_key)
            try:
                table_service.insert_entity(table_name, entity, timeout=timeout)
            except azure.core.exceptions.ResourceExistsError:
                table_service.update_entity(table_name, entity, timeout=timeout)
            except (TypeError, ValueError) as e:
                print(e)
                print('Entity badly formatted.')
            except azure.common.AzureHttpError:
                print('Azure HTTP error.')
        return True
    
    def _set_entity(self,
                    entity: dict,
                    partition_key: str,
                    row_key: str) -> dict:
        """
        :param entity: [description]
        :param partition_key: [description]
        :param row_key: [description]
        :return: [description]
        """
        if partition_key.startswith('$.'):
            partition_key_val = entity.get(partition_key[2:]) 
        else:
            partition_key_val = partition_key
        if row_key.startswith('$.'):
            row_key_val = entity.get(row_key[2:]) 
        else:
            row_key_val = row_key
        #
        entity['PartitionKey'] = partition_key_val
        entity['RowKey'] = row_key_val
        #
        for k, v in entity.items():
            if type(v) not in (str, float, int, bool, datetime):
                entity[k] = json.dumps(v)
        return entity


class LoaderAzureDataLake(BaseLoader):
    
    def __init__(self,
                 metadata: dict = None):
        super().__init__(metadata)
        
    @override(BaseLoader)
    def load(self,
             data_blocks: Iterable[BaseDataBlock],
             storage_account_name: str,
             storage_account_key: str,
             file_system_name: str,
             directory: str = '',
             overwirte: bool = True,
             append_data_kwargs: dict = None,
             **kwargs) -> bool:
        """
        :param storage_account_name:
        :param storage_account_key:
        :param data_blocks: the list of BaseDataBlock to load into the blob storage.
        :param file_system_name: the file system name where the blob is located.
        :param directory: the directory where the blob is located.
        :param overwrite: overwrite blob's previous content, defaults to True
        :return: boolean indicating a succesful load.
        """
        append_data_kwargs = append_data_kwargs or dict()
        storage_account_url = 'https://%s.dfs.core.windows.net/' % storage_account_name
        data_lake_service_client = DataLakeServiceClient(storage_account_url, credential=storage_account_key)
        file_system_client = data_lake_service_client.get_file_system_client(file_system_name)
        for data_block in data_blocks:
            file_path = directory + data_block.name
            file_client = file_system_client.get_file_client(file_path)
            data = data_block.__str__()
            file_client.append_data(data, offset=0, **append_data_kwargs)
            file_client.flush_data(offset=len(data))
        return True


class LoaderAzureFileShare(BaseLoader):
    """
    """
    def __init__(self,
                 metadata: dict = None):
        super().__init__(metadata)

    def load(self,
             data_blocks: Iterable[BaseDataBlock],
             storage_account_name: str,
             storage_account_key: str,
             share_name: str,
             directory: str,
             format: str = 'text',
             overwrite: bool = False,
             **kwargs) -> bool:
        """
        :param data_blocks: [description]
        :type data_blocks: List[BaseDataBlock]
        :param storage_account_name: [description]
        :type storage_account_name: str
        :param storage_account_key: [description]
        :type storage_account_key: str
        :param share_name: [description]
        :type share_name: str
        :param directory: [description]
        :type directory: str
        :param write_type: either 'text' or 'bytes', defaults to 'text'
        :type write_type: str, optional
        :return: [description]
        :rtype: bool
        """
        file_service = FileService(storage_account_name, storage_account_key)
        for data_block in data_blocks:
            if format == 'text':
                file_service.create_file_from_text(share_name, directory, data_block.name,
                                                   data_block.__str__())
            elif format == 'bytes':
                file_service.create_file_from_bytes(share_name, directory, data_block.name,
                                                    data_block.records)
            else:
                raise ValueError(f"format must be either 'text' or 'bytes', not '{format}'")



from .data_block import(
    BaseDataBlock,
    DataBlock,
    BaseRelationalDataBlock,
    RelationalDataBlock,
    JSONDataBlock
)

from .pandas_data_block import PandasDataBlock
from .dask_data_block import DaskDataBlock



from __future__ import annotations
from typing import (
    List,
    Dict,
    NamedTuple
)
import email
import base64
from ._base_message import BaseMessage
from .._base_object import _BaseObject


def make_message(**kwargs) -> BaseMessage:
    if 'payload' in kwargs:
        if 'body' in kwargs['payload']:
            return FullMessage(**kwargs)
        else:
            return MetadataMessage(**kwargs)
    elif 'raw' in kwargs:
        return RawMessage(**kwargs)
    else:
        return MinimalMessage(**kwargs)
    

class MessageWithHeaders(BaseMessage):
    
    def get_header(self, header: str) -> str:
        for header_data in self._payload.headers:
            if header_data['name'] == header:
                return header_data['value']
        return str()
    
    @property
    def from_(self) -> str:
        return self.get_header('From')
    
    @property
    def to(self) -> str:
        return self.get_header('To')
    
    @property
    def subject(self) -> str:
        return self.get_header('Subject')


class FullMessage(MessageWithHeaders):

    def __init__(self,
                 payload: Dict[str, object] = None,
                 **kwargs):
        super().__init__(**kwargs)
        self._payload = FullMessage.Payload(**payload)    

    @property
    def payload(self) -> FullMessage.Payload:
        return self._payload
    
    #@override(BaseMessage)
    @property
    def email_message(self) -> email.message.Message:
        """
        """
        msg_bytes = None
        #
        if 'data' in self._payload.body:
            msg_bytes = base64.urlsafe_b64decode(self._payload.body['data'].encode('utf-8'))
        elif self._payload.parts:
            part = FullMessage.get_part_by_mime_type(self._payload.parts)
            if part:
                msg_bytes = base64.urlsafe_b64decode(part['body']['data'].encode('utf-8'))
        #
        return email.message_from_bytes(msg_bytes) if msg_bytes else email.message.Message()
    
    @staticmethod
    def get_part_by_mime_type(parts: List[dict],
                              mime_type='text/plain'):
        for part in parts:
            if part['mimeType'] == mime_type:
                return part
            if part['mimeType'].startswith('multipart'):
                result = FullMessage.get_part_by_mime_type(part['parts'], mime_type)
                if result:
                    return result
        return dict()

    def to_string(self) -> str:
        for part in self.email_message.walk():
            if part.get_content_type() == 'text/plain':
                payload = part.get_payload(decode=True)
                if payload:
                    return part.get_payload(decode=True).decode('utf-8')
        return str()
    
    class Payload(NamedTuple):
        partId: str
        mimeType: str
        filename: str
        headers: List[Dict[str, str]]
        body: Dict[str, object] = None
        parts: List[Dict[str, object]] = None


class MetadataMessage(MessageWithHeaders):

    def __init__(self,
                 payload: Dict[str, object] = None,
                 **kwargs):
        super().__init__(**kwargs)
        self._payload = MetadataMessage.Payload(**payload)
    
    @property
    def payload(self) -> MetadataMessage.Payload:
        return self._payload
    
    class Payload(NamedTuple):
        mimeType: str
        headers: List[Dict[str, str]]


class RawMessage(BaseMessage):

    def __init__(self,
                 raw: str,
                 **kwargs):
        super().__init__(**kwargs)
        self._raw = raw

    #@override(BaseMessage)
    @property
    def email_message(self) -> email.message.Message:
        """
        """
        msg_bytes = base64.urlsafe_b64decode(self._raw.encode('utf-8'))
        return email.message_from_bytes(msg_bytes)

    def to_string(self) -> str:
        for part in self.email_message.walk():
            if part.get_content_type() == 'text/plain':
                payload = part.get_payload(decode=True)
                if payload:
                    return part.get_payload(decode=True).decode('utf-8')
        return str()
    
    
class MinimalMessage(BaseMessage):
    pass



from typing import (
    List,
    Tuple,
    Any,
    Union
)
import pyodbc
from npyetl.extraction import BaseExtractor
from npyetl.data import (
    RelationalDataBlock,
    PandasDataBlock
)
from npyetl.internals.decorators import override


class ExtractorSQL(BaseExtractor):
    """
    An Extractor class for SQL Databases.
    """
    @override(BaseExtractor)
    def extract(self,
                connection_string: str,
                query: str,
                query_params: Union[List[Any], Tuple[Any, ...]] = ()) -> List[RelationalDataBlock]:
        """
        """
        data_base = pyodbc.connect(connection_string)
        return PandasDataBlock.from_sql(data_base, query % query_params)


class ExtractorSQLFunction(ExtractorSQL):
    """
    An ExtractorSQL child class for SQL Database Functions.
    """
    FUNCTION_QUERY = "SELECT * FROM %s(%s)"
    
    @override(ExtractorSQL)
    def extract(self,
                connection_string: str,
                function_name: str,
                function_parameters: dict) -> List[RelationalDataBlock]:
        """
        """
        query = self.FUNCTION_QUERY % (function_name, tuple(function_parameters.values()))
        return super(ExtractorSQLFunction, self).extract(connection_string, query)



from typing import (
    List,
    Generator,
    Type
)
from googleapiclient.discovery import build
from datetime import datetime
import numpy as np
from npyetl.collab.gmail_.gmail.objects import (
    make_message,
    BaseMessage,
    FullMessage,
    Thread,
    Label,
    Draft,
    History
)


class Api:

    def __init__(self,
                 credentials: dict,
                 version: str = 'v1'):
        self._service = build('gmail', version, credentials=credentials)

    def get_labels(self) -> List[Label]:
        """
        Returns all labels in the user's mailbox.
        
        :returns:
            A list of :class:`gmail.objects.Label` instances.
        """
        return self._service.users().labels().list(userId='me').execute()
    
    def get_message(self,
                    id: str,
                    user_id: str = 'me',
                    format: str = 'raw') -> BaseMessage:
        """
        Returns the specified message.
        
        :param id:
            The ID of the message to retrieve.
        :param user_id:
            The user's email address. Special value 'me'.
        :param format:
            The format to return the message in.
        :return:
            A list of :class:`gmail.objects.BaseMessage` instances.
        """
        if format not in ('full','metadata','minimal','raw'):
            raise ValueError("Parameter 'format' must be one of 'full','metadata','minimal','raw'")
        response = self._service.users().messages().get(userId=user_id,id=id,format=format).execute()
        return make_message(**response)

    def get_messages(self,
                     user_id: str = 'me',
                     include_spam_trash: bool = False,
                     label_ids: List[str] = None,
                     max_results: int = None,
                     query: str = None,
                     format: str = 'raw') -> Generator[BaseMessage, None, None]:
        """
        Returns the messages in the user's mailbox.
        
        :param user_id:
            The user's email address. The special value 'me' can be used to indicate the 
            authenticated user, defaults to 'me'
        :param include_spam_trash:
            Include messages from **SPAM** and **TRASH** in the results, defaults to False
        :param label_ids:
            Only return messages with labels that match all of the specified label IDs.
        :param max_results:
            Maximum number of messages to return.
        :param query:
            Only return messages matching the specified query. Supports the same query format
            as the Gmail search box. For example, **"from:someuser@example.com
            rfc822msgid:<somemsgid@example.com> is:unread"**. Parameter cannot be used when
            accessing the api using the gmail.metadata scope.
        :param format:
            The format to return the message in. Accepted values are:
            * "full": Returns the full email message data with body content parsed in the
                      payload field; the raw field is not used. (default)
            * "metadata": Returns only email message ID, labels, and email headers.
            * "minimal": Returns only email message ID and labels; does not return the email
                         headers, body, or payload.
            * "raw": Returns the full email message data with body content in the raw field
                     as a base64url encoded string; the payload field is not returned.
        :return:
            A list of :class:`gmail.objects.BaseMessage` instances.
        """
        list_params = {'userId':user_id, 'includeSpamTrash':include_spam_trash, 'labelIds':label_ids,
                       'maxResults':max_results, 'q':query}
        #
        response = self._service.users().messages().list(**list_params).execute()
        count = response['resultSizeEstimate']
        #
        if 'messages' in response:
            for message in response['messages']:
                yield self.get_message(message['id'], user_id, format)
        #
        while 'nextPageToken' in response and (not max_results or count < max_results):
            page_token = response['nextPageToken']
            response = self._service.users().messages().list(pageToken=page_token,**list_params).execute()
            count += response['resultSizeEstimate']
            if 'messages' in response:
                for message in response['messages']:
                    yield self.get_message(message['id'], user_id, format)

    def get_thread(self,
                   id: str,
                   user_id: str = 'me') -> Thread:
        """
        """
        return Thread.from_dict(self._service.users().threads().get(userId=user_id, id=id).execute())

    def get_threads(self,
                    user_id: str = 'me',
                    query: str = None,
                    max_results: int = np.inf) -> Generator[Thread, None, None]:
        """
        """
        response = self._service.users().threads().list(userId=user_id, q=query,
                                                        maxResults=max_results).execute()
        count = response['resultSizeEstimate']
        #
        if 'threads' in response:
            for thread in response['threads']:
                yield self.get_thread(thread['id'], user_id)
        #
        while 'nextPageToken' in response and count < max_results:
            page_token = response['nextPageToken']
            response = self._service.users().threads().list(userId=user_id, pageToken=page_token,
                                                            q=query, maxResults=max_results).execute()
            count += response['resultSizeEstimate']
            for thread in response['threads']:
                yield self.get_thread(thread['id'], user_id)



from __future__ import annotations
from typing import (
    NamedTuple,
    Dict
)
from ._base_object import (
    _BaseObject
)

class Label(_BaseObject):
    
    def __init__(self,
                 id: str,
                 name: str,
                 messageListVisibility: str,
                 labelListVisibility: str,
                 type: str,
                 messagesTotal: int,
                 messagesUnread: int,
                 threadsTotal: int,
                 threadsUnread: int,
                 color: Dict[str,str]):
        super().__init__(id)
        self._name = name
        self._message_list_visibility = messageListVisibility
        self._label_list_visibility = labelListVisibility
        self._type = type
        self._messages_total = messagesTotal
        self._messages_unread = messagesUnread
        self._threads_total = threadsTotal
        self._threads_unread = threadsUnread
        self._color = _Color(**color)

    @property
    def name(self) -> str:
        return self._name
        
    @property
    def message_list_visibility(self) -> str:
        return self._message_list_visibility
        
    @property
    def label_list_visibility(self) -> str:
        return self._label_list_visibility
        
    @property
    def type(self) -> str:
        return self._type
        
    @property
    def messages_total(self) -> int:
        return self._messages_total
        
    @property
    def messages_unread(self) -> int:
        return self._messages_unread
        
    @property
    def threads_total(self) -> int:
        return self._threads_total
        
    @property
    def threads_unread(self) -> int:
        return self._threads_unread
        
    @property
    def color(self) -> _Color:
        return self._color

        
class _Color(NamedTuple):
    """
    """
    textColor: str
    backgorundColor: str
    


from typing import (
    Iterable
)
from abc import(
    ABC,
    abstractmethod
)
from npyetl.data import BaseDataBlock


class BaseLoader(ABC):
    
    def __init__(self,
                 metadata: dict = None):
        self._metadata = metadata or dict()
        
    @abstractmethod
    def load(self,
             data_blocks: Iterable[BaseDataBlock],
             **kwargs) -> bool:
        raise NotImplementedError



from typing import (
    List,
    Dict
)
import time
from npyetl.pipeline.logger_loader import BaseLoggerLoader


class PipelineLogger:
    
    def __init__(self,
                 pipeline_name: str,
                 loader: BaseLoggerLoader = None):
        self._pipeline_name = pipeline_name
        self._loader = loader
        self._logs = list()
        
    def log(self, log: Dict[str, object]):
        self._logs.append(log)
        if self._loader:
            self._loader.load(log)

    def log_start(self, **kwargs):
        self.log(self._make_log('start', **kwargs))
    
    def log_end(self, **kwargs):
        self.log(self._make_log('end', **kwargs))

    def _make_log(self,
                  status: str,
                  **kwargs) -> Dict[str, str]:
        """
        [summary]
        :param status: [description]
        :return: [description]
        """
        log = { 'pipeline_name': self._pipeline_name, 'status':status }
        log.update(kwargs)
        return log



from typing import (
    Union,
    Iterable,
    Dict
)
import datetime
import pickle
from google.oauth2.credentials import Credentials
from google.auth.transport.requests import Request
import importlib
#
from npyetl.collab.gmail_ import gmail
from npyetl.extraction.extractor import BaseExtractor
from npyetl.load.loader import BaseLoader
from npyetl.data import DataBlock
from npyetl.internals import override


def my_import(name):
    components = name.split('.')
    mod = __import__(components[0])
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod

def get_and_update_credentials(extractor_name: str,
                               loader_name: str,
                               params: dict) -> Credentials:
    """[summary]
    :param reader: [description]
    :type reader: BaseExtractor
    :param writer: [description]
    :type writer: BaseLoader
    :param params: [description]
    :type params: dict
    :return: [description]
    :rtype: Credentials
    """
    extractor = my_import(extractor_name)
    loader = my_import(loader_name)
    creds_bytes = next(extractor().extract(**params)).records
    creds = pickle.loads(creds_bytes)
    if not creds.valid and creds.expired and creds.refresh_token:
        creds.refresh(Request())
        loader().load([DataBlock(pickle.dumps(creds))], **params)
    return creds

def make_query(start_time: Union[datetime.datetime, int, float],
               end_time: Union[datetime.datetime, int, float],
               query: str = '',) -> str:
    """[summary]
    :param start_time: [description]
    :param end_time: [description]
    :param query: [description], defaults to ''
    :return: [description]
    """
    if isinstance(start_time, datetime.datetime):
        start_time = start_time.timestamp()
    if isinstance(end_time, datetime.datetime):
        end_time = end_time.timestamp()            
    return f"after:{int(start_time)} before:{int(end_time)} " + query


class ExtractorGmailMessage(BaseExtractor):

    @override(BaseExtractor)
    def extract(self,
                credentials: Dict[str, object],
                start_time: Union[datetime.datetime, int, float],
                end_time: Union[datetime.datetime, int, float],
                query: str = '',
                api_version: str = 'v1') -> Iterable[DataBlock]:
        """
        [summary]
        :param credentials: [description]
        :param start_time: [description]
        :param end_time: [description]
        :param query: [description], defaults to ''
        :param api_version: [description], defaults to 1
        :return: [description]
        """
        credentials = get_and_update_credentials(**credentials)
        api = gmail.Api(credentials, api_version)
        query = make_query(start_time, end_time, query)
        print(f"Extracting messages for query {query}")
        messages = api.get_messages(query=query, format='full')
        for message in messages:
            yield DataBlock(message, name=message.id) 
        
        
class ExtractorGmailThread(BaseExtractor):

    @override(BaseExtractor)
    def extract(self,
                credentials: Dict[str, Union[str, dict]],
                start_time: Union[datetime.datetime, float],
                end_time: Union[datetime.datetime, float],
                query: str = '',
                api_version: str = 'v1') -> Iterable[DataBlock]:
        """[summary]
        :param credentials: [description]
        :type credentials: dict
        :param start_time: [description]
        :type start_time: Union[datetime.datetime, float]
        :param end_time: [description]
        :type end_time: Union[datetime.datetime, float]
        :param query: [description], defaults to ''
        :type query: str, optional
        :param api_version: [description], defaults to 1
        :type api_version: int, optional
        :return: [description]
        :rtype: List[DataBlock]
        """
        credentials = get_and_update_credentials(**credentials)
        api = gmail.Api(credentials, api_version)
        query = make_query(start_time, end_time, query)
        threads = api.get_threads(query=query)
        for thread in threads:
            yield DataBlock(thread, name=thread.id) 



from __future__ import annotations
from abc import (
    ABC
)
import json


class _BaseObject(ABC):
    
    def __init__(self,
                 id: str):
        self._id = id
        
    @classmethod
    def from_dict(cls, dictionary: dict) -> _BaseObject:
        return cls(**dictionary)
    
    @classmethod
    def from_json(cls, json_content: str):
        return cls.from_dict(json.loads(json_content))
    
    @property
    def id(self) -> str:
        return self._id



from typing import (
    Iterable
)
from abc import (
    ABC,
    abstractmethod
)
from npyetl.data import BaseDataBlock


class BaseExtractor(ABC):
    
    def __init__(self,
                 metadata: dict = None):
        self._metadata = metadata or dict()
    
    @abstractmethod
    def extract(self) -> Iterable[BaseDataBlock]:
        raise NotImplementedError



from typing import (
    List,
    Iterable
)
import os
import json
import requests
#
from npyetl.extraction import BaseExtractor
from npyetl.data import (
    JSONDataBlock
)
from npyetl.internals.decorators import override


class ExtractorJSON(BaseExtractor):
    """
    TODO: map flexible input to adequate child class.
    """
    @override(BaseExtractor)
    def extract(self,
                **kwargs) -> Iterable[JSONDataBlock]:
        raise NotImplementedError


class ExtractorLocalJSON(ExtractorJSON):
    """
    An ExtractorJSON child class for local JSON files.
    """
    @override(ExtractorJSON)  
    def extract(self,
                directory: str,
                matches: List[str] = [],
                headers: bool = True) -> Iterable[JSONDataBlock]:
        """
        """
        for file_name in [ f for f in os.listdir(directory) if f.endswith('.json') ]:
            if not matches or any([m in file_name for m in matches]): 
                file = open(directory + file_name, 'r')
                records = json.load(file)
                yield JSONDataBlock(records, name=file_name)
        return


class ExtractorUrlJSON(ExtractorJSON):
    """
    An ExtractorJSON child class for urls containing JSON files.
    """
    @override(ExtractorJSON)
    def extract(self,
                file_url: str,
                **kwargs) -> Iterable[JSONDataBlock]:
        """
        """
        records = json.loads(requests.get(file_url).content)
        yield JSONDataBlock(records)



from typing import (
    Iterable,
    Optional
)
from npyetl.data import RelationalDataBlock
from npyetl.load import BaseLoader
from npyetl.internals.decorators import override


class LoaderCSV(BaseLoader):
    """
    TODO: map flexible input to adequate child class.
    """
    @override(BaseLoader)
    def load(self,
             **kwargs) -> bool:
        raise NotImplementedError


class LoaderLocalCSV(LoaderCSV):
    """
    Loader class to a local CSV file
    """
    
    @override(BaseLoader)
    def load(self,
             data_blocks: Iterable[RelationalDataBlock],
             directory: str = '') -> bool:
        for data_block in data_blocks:
            file = open(directory + data_block.name +'.csv', 'w')
            file.write(data_block.__str__())
            file.close()
        return True



from typing import (
    Dict
)
from ._base_object import (
    _BaseObject
)
from .message import (
    BaseMessage,
    make_message
)


class Draft(_BaseObject):
    
    def __init__(self,
                 id: str,
                 message: Dict[str, object]):
        super().__init__(id)
        self._message = make_message(**message)
    
    @property
    def message(self) -> BaseMessage:
        return self._message



from typing import (
    Optional,
    Iterable
)
#
from azure.storage.blob import BlobServiceClient
from azure.cosmosdb.table import TableService
from azure.storage.queue import QueueService
from azure.storage.file import FileService
from azure.storage.file.models import (
    File,
    Share,
    Directory
)
#
from npyetl.internals.decorators import override
from npyetl.extraction import BaseExtractor
from npyetl.data import (
    BaseDataBlock,
    DataBlock,
    JSONDataBlock
)


class ExtractorAzureBlob(BaseExtractor):
    """
    Extractor for Azure Blob Storage blob files.
    """        
    def extract(self,
                storage_account_name: str,
                storage_account_key: str,
                container_name: str,
                directory: str = '',
                blob_name: str = None,
                **kwargs):
        """[summary]
        :param storage_account_name: [description]
        :type storage_account_name: str
        :param storage_account_key: [description]
        :type storage_account_key: str
        :param container_name: [description]
        :type container_name: str
        :param blob_name: [description]
        :type blob_name: str
        :param directory: [description], defaults to ''
        :type directory: str, optional
        """
        blob_storage_url = 'https://%s.blob.core.windows.net/' % storage_account_name
        blob_service_client = BlobServiceClient(blob_storage_url, credential=storage_account_key)
        container_client = blob_service_client.get_container_client(container_name)
        if blob_name:
            blob_client = blob_service_client.get_blob_client(container_name, directory+blob_name)
            download_stream = blob_client.download_blob()
            yield DataBlock(download_stream.readall(), name=blob_name)
        else:
            container_client = blob_service_client.get_container_client(container_name)
            for blob in container_client.list_blobs(name_starts_with=directory):
                blob_client = container_client.get_blob_client(blob.name)
                download_stream = blob_client.download_blob()
                yield DataBlock(download_stream.readall(), name=blob_name)
        return


class ExtractorAzureTable(BaseExtractor):
    """
    """
    def extract(self,
                storage_account_name: str,
                storage_account_key: str,
                table_name: str,
                filter: Optional[str] = None,
                select: Optional[str] = None,
                num_results: Optional[int] = None,
                **kwargs):
        """
        """
        table_service = TableService(storage_account_name, storage_account_key)
        for entity in table_service.query_entities(table_name, filter, select, num_results):
            yield JSONDataBlock(entity)
            

class ExtractorAzureQueue(BaseExtractor):
    """
    """
    def extract(self,
                storage_account_name: str,
                storage_account_key: str,
                queue_name: str,
                pop: bool = True,
                num_messages: Optional[int] = None,
                visibility_timeout: Optional[int] = None,
                timeout: Optional[int] = None,
                **kwargs):
        """
        """
        queue_service = QueueService()
        for message in queue_service.get_messages(queue_name, num_messages, visibility_timeout, timeout):
            yield DataBlock(message.content)
            if pop:
                queue_service.delete_message(queue_name, message.id, message.pop_receipt, timeout)


class ExtractorAzureFileShare(BaseExtractor):
    """
    Extractor for Azure File Share files.
    """        
    def extract(self,
                storage_account_name: str,
                storage_account_key: str,
                share_name: str,
                directory: str,
                file_name: Optional[str] = None,
                format: str = 'text',
                **kwargs) -> Iterable[BaseDataBlock]:
        """
        [summary]
        :param storage_account_name: [description]
        :type storage_account_name: str
        :param storage_account_key: [description]
        :type storage_account_key: str
        :param share_name: [description]
        :type share_name: str
        :param directory: [description]
        :type directory: str
        :param file_name: [description]
        :type file_name: str
        :param return_type: either 'text' or 'bytes', defaults to 'text'
        :type return_type: type, optional
        :return: [description]
        :rtype: BaseDataBlock
        """
        file_service = FileService(storage_account_name, storage_account_key)
        get_func = getattr(file_service, 'get_file_to_'+format)
        if file_name:
            print(share_name, directory, file_name)
            yield DataBlock(get_func(share_name, directory, file_name).content, name=file_name)
        #
        else:
            for file_or_dir in file_service.list_directories_and_files(share_name, directory):
                if isinstance(file_or_dir, File):
                    file_name = file_or_dir.name
                    yield DataBlock(get_func(share_name, directory, file_name).content, name=file_name)
        return
        


from npyetl.internals.decorators import(
    override
)
from abc import(
    ABC,
    abstractmethod,
    abstractproperty
)
from typing import(
    List,
    Dict,
    Union,
    Generator,
    Any
)
import json
import uuid


class BaseDataBlock(ABC):
    """
    """
    @property
    @abstractmethod
    def records(self) -> object:
        return NotImplementedError

    @abstractmethod
    def __str__(self) -> str:
        return NotImplementedError

    def __len__(self) -> int:
        return len(self.records)
    
    def __iter__(self) -> Generator[str, None, None]:
        return (record for record in self.records)
    
    @abstractproperty
    def metadata(self) -> Dict[str, object]:
        return NotImplementedError

    @abstractproperty
    def name(self) -> str:
        return NotImplementedError


class DataBlock(BaseDataBlock):
    """
    """
    def __init__(self,
                 records: object,
                 metadata: dict = None,
                 name: str = None):
        self._records = records
        self._metadata = metadata or dict()
        self._name = name
    
    @property
    @override(BaseDataBlock)
    def records(self) -> object:
        return self._records
    
    @records.setter
    def records(self, new_value: object):
        self._records = new_value

    @property
    @override(BaseDataBlock)
    def metadata(self) -> dict:
        return self._metadata

    @property
    @override(BaseDataBlock)
    def name(self) -> str:
        return self._name or uuid.uuid4().__str__()
    
    @name.setter
    def name(self, new_name: str):
        self._name = new_name

    @override(BaseDataBlock)
    def __str__(self):
        return '\n'.join([ record.__str__() for record in self.records ])
    


class BaseRelationalDataBlock(DataBlock):
    """
    """
    @override(BaseDataBlock)
    def __str__(self) -> str:
        string = ''
        if self.columns:
            string += ','.join(self.columns) + '\n'
        string += '\n'.join([ ','.join(record) for record in self.records ])
        return string

    def __getitem__(self,
                    key: str) -> list:
        """Gets one column's data."""
        colIx = self._columns.index(key) 
        if colIx == -1:
            raise KeyError(f"'{key}' is not a column.")
        return [ record[colIx] for record in self.records ]


class RelationalDataBlock(BaseRelationalDataBlock):
    """
    """
    def __init__(self,
                 records: List[List[Any]],
                 columns: List[str] = None,
                 types: Dict[str, type] = None,
                 metadata: dict = None,
                 **kwargs):
        """
        :param columns: the column names for the records.
        :param records: the actual data rows.
        :param types: the types for each column.
        :param metadata: the metadata of the data block.
        """
        super().__init__(records, metadata, **kwargs)
        self._columns = columns or list()
        self._types = types or dict()
    
    @property
    def columns(self) -> List[str]:
        return self._columns
    
    @property
    def types(self) -> Dict[str, type]:
        return self._types


class JSONDataBlock(DataBlock):
    """
    """
    def __init__(self,
                 records: object,
                 schema: dict = None,
                 metadata: dict = None,
                 **kwargs):
        """
        :param records: the actual data rows.
        :param schema: the JSON schema for the rows.
        :param metadata: the metadata of the data block.
        """
        if type(records) == str:
            records = json.loads(records)
        else:
            try:
                json.dumps(records)
            except:
                raise ValueError("Invalid JSON object")
        super().__init__(records, metadata, **kwargs)
    
    @override(BaseDataBlock)
    def __str__(self) -> str:
        return json.dumps(self._records)



from typing import (
    List,
    Dict
)
from ._base_object import (
    _BaseObject
)
from .message import (
    make_message,
    BaseMessage
)


class Thread(_BaseObject):
    
    def __init__(self,
                 id: str,
                 historyId: str,
                 messages: List[Dict[str, object]]):
        super().__init__(id)
        self._history_id = historyId
        self._messages = [ make_message(**message) for message in messages ]
    
    @property
    def history_id(self) -> str:
        return self._history_id

    @property
    def messages(self) -> List[BaseMessage]:
        return self._messages



from abc import(
    ABC,
    abstractmethod
)
from typing import(
    List,
    Iterable,
    Dict,
    Callable,
    Any,
    Optional,
    Union
)
#
from npyetl.data import BaseDataBlock
from npyetl.internals.decorators import override

#
TransformFunction = Callable[[BaseDataBlock, Dict[str, object]], BaseDataBlock]


class BaseTransformer(ABC):
    
    def __init__(self,
                 metadata: dict = None):
        self._metadata = metadata or dict()
        
    @abstractmethod
    def transform(self,
                  input_data_blocks: Iterable[BaseDataBlock],
                  **kwargs) -> Iterable[BaseDataBlock]:
        raise NotImplementedError


class Transformer(BaseTransformer):
    
    def __init__(self,
                 transform_functions: List[TransformFunction],
                 metadata: dict = None):
        super().__init__(metadata)
        self._transform_functions = transform_functions
        
    @override(BaseTransformer)
    def transform(self,
                  input_data_blocks: Iterable[BaseDataBlock],
                  transform_functions_params: List[Dict[str, object]] = None) -> Iterable[BaseDataBlock]:
        """
        """
        if not transform_functions_params:
            transform_functions_params = [dict()]*len(self._transform_functions)
        assert len(transform_functions_params) == len(self._transform_functions)
        #
        for input_data_block in input_data_blocks:
            transformed_data_block = input_data_block
            for i, transform_function in enumerate(self._transform_functions):
                transformed_data_block = transform_function(transformed_data_block, **transform_functions_params[i])
            yield transformed_data_block
        return



from typing import(
    List,
    Dict,
    Iterable,
    Optional
)
#
import multiprocessing as mp 
import datetime
import time
import json
import uuid
#
from npyetl.data import BaseDataBlock
from npyetl.extraction import BaseExtractor
from npyetl.transformation import BaseTransformer
from npyetl.load import BaseLoader
from npyetl.pipeline.logger import (
    PipelineLogger,
    BaseLoggerLoader
)


class Pipeline:
    
    def __init__(self,
                 extractor: BaseExtractor,
                 loader: BaseLoader,
                 transformer: BaseTransformer = None,
                 logger_loader: BaseLoggerLoader = None,
                 pipeline_parameters: Optional[Dict[str, object]] = None,
                 name: str = None,
                 metadata: dict = None):
        """[summary]
        :param extractor: [description]
        :type extractor: BaseExtractor
        :param loader: [description]
        :type loader: BaseLoader
        :param transformer: [description], defaults to None
        :type transformer: BaseTransformer, optional
        :param logger: [description], defaults to None
        :type logger: PipelineLogger, optional
        :param pipeline_params: [description], defaults to None
        :type pipeline_params: Optional[Dict[str, object]], optional
        :param name: [description], defaults to None
        :type name: str, optional
        :param metadata: [description], defaults to None
        :type metadata: dict, optional
        """
        self._extractor = extractor
        self._transformer = transformer
        self._loader = loader
        self._params = pipeline_parameters or dict()
        self._name = name
        self._metadata = metadata or dict()
        #
        self._logger = PipelineLogger(self._name, logger_loader) if logger_loader else None
        
    @classmethod
    def from_dict(cls, dictionary: dict) -> 'Pipeline':
        return cls(**dictionary)

    @classmethod
    def from_json(cls, json_content: str) -> 'Pipeline':
        return cls.from_dict(json.loads(json_content))

    @property
    def name(self) -> str:
        return self._name or self.__class__.__name__
    
    @property
    def id(self) -> str:
        return uuid.uuid4().__str__()

    @property
    def metadata(self) -> Optional[dict]:
        return self._metadata

    @property 
    def state(self) -> str:
        return self._state

    @property
    def logger(self) -> Optional[PipelineLogger]:
        return self._logger

    @property
    def params(self) -> Optional[Dict[str,object]]:
        return self._params

    ##################################################################################
    def execute(self) -> Dict[str,object]:
        """
        """
        start = time.time()
        self._log_start()
        succesful = self._load(self._transform(self._extract()))
        stats = {'execution_time': time.time()-start,
                 'pipeline_name': self.name,
                 'succesful': succesful}
        self._log_end(**stats)
        return stats

    def _extract(self) -> Iterable[BaseDataBlock]:
        extract_params = self._params.get('extraction', dict())
        yield from self._extractor.extract(**extract_params)
        return

    def _transform(self, data_blocks: Iterable[BaseDataBlock]) -> Iterable[BaseDataBlock]:
        if self._transformer:
            transformation_params = self._params.get('transformation', dict())
            yield from self._transformer.transform(data_blocks, **transformation_params)
            return
        yield from data_blocks
        return
        

    def _load(self, data_blocks: Iterable[BaseDataBlock]) -> bool:
        load_params = self._params.get('load', dict())
        load_status = self._loader.load(data_blocks, **load_params)
        return load_status
    
    ################################################################################
    def _log_start(self, **kwargs):
        if self._logger:
            self._logger.log_start(**kwargs)

    def _log_end(self, **kwargs):
        if self._logger:
            self._logger.log_end(**kwargs)


class PipelinePool:

    def __init__(self,
                 pipelines: List[Pipeline],
                 max_threads: int = 32,
                 metadata: dict = None):
        self._pipelines = pipelines
        self._num_of_pipelines = len(pipelines)
        self._max_threads = max_threads
        self._metadata = metadata

    @property
    def pipelines(self) -> List[Pipeline]:
        return self._pipelines

    @property
    def loggers(self) -> Dict[str, Optional[PipelineLogger]]:
        return { pipeline.id: pipeline.logger for pipeline in self._pipelines }

    @property
    def metadata(self) -> Optional[dict]:
        return self._metadata

    def execute(self) -> Iterable[Dict[str, object]]:
        with mp.Pool(min(self._max_threads, self._num_of_pipelines)) as pool:
            yield from list(pool.map(lambda p: p.execute(), self._pipelines))



from .._base_object import _BaseObject


class Attachment(_BaseObject):
    
    def __init__(self,
                 id: str,
                 size: int,
                 data: bytes):
        super().__init__(id)
        self._size = size
        self._data = data
    
    @property
    def size(self) -> int:
        return self._size
    
    @property
    def data(self) -> bytes:
        return self._data



from typing import (
    List,
    Iterable
)
import os
import csv
from npyetl.extraction import BaseExtractor
from npyetl.data import (
    RelationalDataBlock,
    PandasDataBlock
)
from npyetl.internals.decorators import override


class ExtractorCSV(BaseExtractor):
    """
    TODO: map flexible input to adequate child class.
    """
    @override(BaseExtractor)
    def extract(self,
                **kwargs) -> Iterable[RelationalDataBlock]:
        raise NotImplementedError


class ExtractorLocalCSV(ExtractorCSV):
    """
    An ExtractorCSV child class for local CSV files.
    """
    @override(ExtractorCSV)  
    def extract(self,
                directory: str,
                matches: List[str] = [],
                headers: bool = True) -> Iterable[RelationalDataBlock]:
        """
        """
        for file_name in [ f for f in os.listdir(directory) if f.endswith('.csv') ]:
            if not matches or any([m in file_name for m in matches]): 
                file = open(directory + file_name, 'r')
                records = list(csv.reader(file))
                name = file_name
                if headers:
                    columns = records[0]
                    records = records[1:]
                    yield RelationalDataBlock(records, columns, name=name)
                else:
                    yield RelationalDataBlock(records, name=name)
        return


class ExtractorUrlCSV(ExtractorCSV):
    """
    An ExtractorCSV child class for urls containing CSV files.
    """
    @override(ExtractorCSV)
    def extract(self,
                file_url: str,
                sep: str = ',',
                **kwargs) -> Iterable[PandasDataBlock]:
        """
        """
        yield PandasDataBlock.read_csv(file_url, sep=sep, **kwargs)



from abc import(
    ABC,
    abstractmethod,
    abstractstaticmethod
)
from typing import(
    List,
    Dict
)
import uuid
import json
from azure.cosmosdb.table.tableservice import TableService
from npyetl.internals import override


class BaseLoggerLoader(ABC):
    
    def load(self, log: Dict[str,object]):
        log = self._format_log(log)
        assert self._validate_log(log)
        self._insert_log(log)

    def _format_log(self, log: Dict[str, object]) -> Dict[str, object]:
        return log

    @abstractmethod
    def _insert_log(self, log: Dict[str, object]):
        return NotImplementedError
    
    @staticmethod
    @abstractstaticmethod
    def _validate_log(log: Dict[str,object]) -> bool:
        raise NotImplementedError


class LogFileLoggerLoader(BaseLoggerLoader):
    
    def __init__(self,
                 file_name: str):
        self._file_name = file_name
        self._file = open(self._file_name, 'w')
        
    @override(BaseLoggerLoader)
    def _insert_log(self, log: Dict[str,object]):
        self._file.write(json.dumps(log))

    @override(BaseLoggerLoader)
    def _validate_log(self, log: Dict[str,object]):
        try:
            json.dumps(log)
            return True
        except TypeError:
            return False


class AzureTableLoggerLoader(BaseLoggerLoader):
    
    REQUIRED_LOG_ATTRS = 'PartitionKey', 'RowKey'
    
    def __init__(self,
                 table_service: TableService,
                 table_name: str):
        self._table_service = table_service
        self._table_name = table_name
    
    @override(BaseLoggerLoader)
    def _insert_log(self, log: Dict[str,object]):
        self._table_service.insert_entity(self._table_name, log)

    @override(BaseLoggerLoader)
    def _validate_log(self, log: Dict[str,object]) -> bool:
        if all([attr in log for attr in self.REQUIRED_LOG_ATTRS]):
            return True
        return False



from typing import(
    Generator,
    List
)
#
import pandas as pd
#
from npyetl.data import BaseRelationalDataBlock
from npyetl.internals.decorators import override


class PandasDataBlock(BaseRelationalDataBlock,
                      pd.DataFrame):

    def __init__(self, *args, **kwargs):
        super(pd.DataFrame, self).__init__(*args, **kwargs)

    @classmethod
    def from_sql(cls, con, query):
        return cls(pd.read_sql(query, con))

    @override(BaseRelationalDataBlock)
    def records(self) -> List[pd.Series]:
        return [ list(record) for _, record in self.iterrows() ]
    
    @override(pd.DataFrame)
    def __iter__(self) -> Generator[pd.Series, None, None]:
        return ( record for _, record in self.iterrows() )



from typing import (
    Iterable,
    Optional
)
import json
from npyetl.data import JSONDataBlock
from npyetl.load import BaseLoader
from npyetl.internals.decorators import override


class LoaderJSON(BaseLoader):
    """
    TODO: map flexible input to adequate child class.
    """
    @override(BaseLoader)
    def load(self,
             data_blocks: Iterable[JSONDataBlock],
             **kwargs) -> bool:
        raise NotImplementedError


class LoaderLocalJSON(LoaderJSON):
    """
    Loader class to a local JSON file
    """
    
    @override(BaseLoader)
    def load(self,
             data_blocks: Iterable[JSONDataBlock],
             directory: str = '') -> bool:
        for data_block in data_blocks:
            file = open(directory + data_block.name +'.json', 'w')
            json.dump(data_block.records, file)
        return True



from setuptools import setup, find_packages
import os
import re

def resolve_requirements(file):
    requirements = []
    with open(file) as f:
        req = f.read().splitlines()
        for r in req:
            if r.startswith("-r"):
                requirements += resolve_requirements(
                    os.path.join(os.path.dirname(file), r.split(" ")[1]))
            else:
                requirements.append(r)
    return requirements


def read_file(file):
    with open(file) as f:
        content = f.read()
    return content


def find_version(file):
    content = read_file(file)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", content,
                              re.M)
    if version_match:
        return version_match.group(1)


requirements = resolve_requirements(os.path.join(os.path.dirname(__file__),
                                                 'requirements.txt'))

readme = read_file(os.path.join(os.path.dirname(__file__), "README.md"))
license = read_file(os.path.join(os.path.dirname(__file__), "LICENSE"))
_version = find_version(os.path.join(os.path.dirname(__file__), "npyetl",
                                           "__init__.py"))


setup(
    name='npyetl',
    version=_version,
    packages=find_packages(),
    url='',
    test_suite="unittest",
    long_description=readme,
    long_description_content_type='text/markdown',
    install_requires=requirements,
    tests_require=["coverage"],
    python_requires=">=3.7",
    author="Miquel Escobar",
    author_email="mec@novicell.es",
    license=license,
)
